#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <signal.h>
#include <X11/Xlib.h>
#include <glib.h>
#include <gtk/gtk.h>

#include "master.h"
#include "drummachine.h"

#include "drummachine_if.h"
#include "none_config.h"


extern double note_rates[96];

extern int bpm, beat;
extern int beat_index;
extern int tick;

int drummachine_calc_note(drummachine * dm);
void dm_restart_beat(drummachine * dm);
void dm_play_once(drummachine * dm);


module_class drummachine_mif =
{
  SEQUENCER,
  0,
  (char *) "drummachine",
  (char *) "DrumMachine",
  NULL,
  (void *) drummachine_new,
  (void *) drummachine_clone,
  NULL,
  NULL,
  (void *) none_config_cb,
  (void *) drummachine_if_new,
  (void *) drummachine_if_close_cb
};

drummachine *
drummachine_new()
{
  drummachine *dm;
  gchar buf[4];
  gint i;

  dm = (drummachine *) malloc(sizeof(drummachine));
  dm->module.class = &drummachine_mif;
  dm->module.on = 0;
  dm->module.nr_inputs = 0;
  dm->module.nr_outputs = MAX_CHANNELS;
  for (i = 0; i < MAX_CHANNELS; i++) {
    dm->module.outputs[i] = malloc(sizeof(channel));
    dm->module.outputs[i]->parms.ch_type = CH_TYPE_SEQUENCE;
    dm->module.outputs[i]->data = (seq_channel *) (&(dm->seq_ch[i][0]));
    snprintf(buf, 4, "%d", i);
    sprintf(dm->module.outputs[i]->u_label, buf);
    dm->module.outputs[i]->module = (module *) dm;
  }
  aube_module_clear_outputs((module *) dm);
  sprintf(dm->module.u_label, "drummachine");

  dm->module.do_tick = (void *) drummachine_calc_note;

  aube_add_module((module *) dm);

  dm->vol = 48;
  dm->pan = 16;
  dm->tune = 128;
  dm->dm_on = 0;
  dm->dmi = 0;

  dm_clear(dm);

  return dm;
}

drummachine *
drummachine_clone(drummachine * odm)
{
  drummachine *dm;
  gint i;
  gchar buf[4];

  dm = (drummachine *) malloc(sizeof(drummachine));
  memcpy(dm, odm, sizeof(drummachine));

  dm->module.nr_inputs = 0;
  dm->module.nr_outputs = 1;
  for (i = 0; i < MAX_CHANNELS; i++) {
    dm->module.outputs[i] = malloc(sizeof(channel));
    dm->module.outputs[i]->parms.ch_type = CH_TYPE_SEQUENCE;
    dm->module.outputs[i]->data = (seq_channel *) (&(dm->seq_ch[i][dm->dmi]));
    dm->module.outputs[i]->module = (module *) dm;
    snprintf(buf, 4, "%d", i);
    sprintf(dm->module.outputs[i]->u_label, buf);
  }
  aube_module_clear_outputs((module *) dm);
  sprintf(dm->module.u_label, "drummachine");

  dm->module.do_tick = (void *) drummachine_calc_note;

  aube_add_module((module *) dm);

  return dm;
}

int 
drummachine_calc_note(drummachine * dm)
{
  gint i;

  dm->bi = beat_index;

  for (i = 0; i < MAX_CHANNELS; i++) {
    dm->module.outputs[i]->data = &(dm->seq_ch[i][dm->dmi]);
#if 0
    dm->seq_ch[i].vol = dm->vol;
    dm->seq_ch[i].accent = 0;
    dm->seq_ch[i].pitch = 128.0;
    dm->seq_ch[i].trigger = 0;
#endif
  }

  dm->bi += tick;
  if (dm->bi >= beat) {
    dm_restart_beat(dm);
  }
  return PROC_SUCCESS;
}

void 
dm_restart_beat(drummachine * dm)
{
  gint i;

  dm->bi = 0;

  dm->dmi++;
  dm->dmi %= DM_LENGTH;

#if 0
  for (i = 0; i < MAX_CHANNELS; i++) {
    dm->seq_ch[i].vol = dm->vol;
    dm->seq_ch[i].accent = 0;
    dm->seq_ch[i].pitch = 128.0;
    dm->seq_ch[i].trigger = (dm->sequence[i][dm->dmi].note > 0);
  }
#endif

}

void 
dm_restart(drummachine * dm)
{
  dm->dmi = 0;
}

void 
dm_shift_left(drummachine * dm)
{
  dm->dmi += DM_LENGTH;
  dm->dmi--;
  dm->dmi %= DM_LENGTH;
}

void 
dm_shift_right(drummachine * dm)
{
  dm->dmi++;
  dm->dmi %= DM_LENGTH;
}

void 
dm_clear(drummachine * dm)
{
  int i, j;

  for (j = 0; j < MAX_CHANNELS; j++) {
    for (i = 0; i < DM_LENGTH; i++) {
	dm->seq_ch[j][i].trigger = 0;
	dm->seq_ch[j][i].pitch = 128.0;
      /* dm->sequence[j][i].note = 0; */
    }
  }
}

void 
dm_chaos(drummachine * dm)
{
  gint i, j;

  for (j = 0; j < MAX_CHANNELS; j++) {
    for (i = 0; i < DM_LENGTH; i++) {
      if ((rand() % DM_LENGTH) < DM_SILENCE) {
	dm->seq_ch[j][i].trigger = 0;
	/* dm->sequence[j][i].note = 0; */
      } else {
	dm->seq_ch[j][i].trigger = 1;
	/* dm->sequence[j][i].note = 128; */
      }
    }
  }
}

void 
dm_play_once(drummachine * dm)
{
#if 0
  int i;

  dm->sequence[0][0].note = 128;
  for (i = 1; i < DM_LENGTH; i++) {
    dm->sequence[0][i].note = 0;
  }
#endif
}

void 
dm_play_44(drummachine * dm)
{
#if 0
  int i;

  for (i = 0; i < DM_LENGTH; i++) {
    if (!(i % 4)) {
      dm->sequence[0][i].note = 128;
    } else {
      dm->sequence[0][i].note = 0;
    }
  }
#endif
}

void 
dm_transpose_u12(drummachine * dm)
{
  dm_transpose(dm, 12);
}

void 
dm_transpose_d12(drummachine * dm)
{
  dm_transpose(dm, -12);
}

void 
dm_transpose_u1(drummachine * dm)
{
  dm_transpose(dm, 1);
}

void 
dm_transpose_d1(drummachine * dm)
{
  dm_transpose(dm, -1);
}

void 
dm_transpose(drummachine * dm, int t)
{
#if 0
  int i, nri;
  int note_rate;

  for (i = 0; i < DM_LENGTH; i++) {
    if (dm->sequence[0][i].note > 0) {
      note_rate = dm->sequence[0][i].note;
      for (nri = 0; nri < 96 && ((int) ((double) 64.0 * note_rates[nri]) <= note_rate); nri++);
      nri += t;
      if (nri < 0)
	nri = 0;
      if (nri >= 95)
	nri = 95;
      dm->sequence[0][i].note = (int) ((double) 64 * note_rates[nri]);
    }
  }
#endif
}
